//@ts-check
"use strict";

/**
 * @enum {number}
 */
const BaddieStates = {
  Idle: 0,
  Wander: 1,
  ChaseTarget: 2,
  AttackTarget: 3,
  Hit: 4
}

/**
@typedef BaddieSpawner
@type {{
  pos: { x: number; y: number; };
  difficulty: number;
  countdown: number;
  interval: { min: number; max: number; } 
}}
*/

/**
@typedef Baddie
@type {{
  baddieDesc: BaddieDesc;
  pos: PIXI.Point;
  collideBox: { width: number; height: number; };
  hitbox: { width: number; height: number; };
  speed: number;
  health: number;

  state: BaddieStates;
  stateTimer: number;
  stateTarget: EntityLike;
  knockbackDirection: PIXI.Point;

  pixiRoot: PIXI.Container;
  bodySpritesheet: Spritesheet;
  bodySprite: PIXI.AnimatedSprite;
  shadowSprite: PIXI.Sprite;
}}
*/

/**
 * @typedef BaddieName
 * @type {keyof typeof baddieDescs}
 */

/**
@typedef BaddieDesc
@type {{
  id: BaddieName;
}}
 */

/**
@type {{
  light: BaddieDesc,
  medium: BaddieDesc,
  strong: BaddieDesc,
  ranged: BaddieDesc,
}}
 */
const baddieDescs = {
  light: { id: "light" },
  medium: { id: "medium" },
  strong: { id: "strong" },
  ranged: { id: "ranged" },
};


let combatCountdown = 0;

/**
 * @param {{ x: number; y: number; }} pos
 * @param {BaddieDesc} baddieDesc
 * @returns Baddie
* */
function createBaddie(pos, baddieDesc) {
  const pixiRoot = new PIXI.Container();
  pixiRoot.scale.set(inverseTileSize);
  inGame.activeWorld.pixiLayers["ground"].entitiesContainer.addChild(pixiRoot);

  pixiRoot.position.set(pos.x, pos.y);

  const bodySpritesheet = getSpritesheet(`inGame/baddies/${baddieDesc.id}/${baddieDesc.id}Body`);
  const bodySprite = new PIXI.AnimatedSprite(bodySpritesheet.animations["idle"], false);
  bodySprite.anchor.copyFrom(bodySpritesheet.anchor);
  pixiRoot.addChild(bodySprite);

  const shadowSprite = new PIXI.Sprite(tex("inGame/shadow"));
  shadowSprite.scale.set(inverseTileSize * 1.2);
  shadowSprite.anchor.set(0.5, 0);
  inGame.activeWorld.pixiShadowsLayer.addChild(shadowSprite);

  /** @type Baddie */
  const baddie = {
    pos: new PIXI.Point(pos.x, pos.y),
    collideBox: { width: 0.8, height: 0.4 },
    hitbox: { width: 1.5, height: 2 },
    baddieDesc,
    speed: 4,
    health: 5,

    state: BaddieStates.Idle,
    stateTimer: 0,
    stateTarget: null,
    knockbackDirection: new PIXI.Point(),

    pixiRoot,
    bodySpritesheet,
    bodySprite,
    shadowSprite
  };

  setBaddieIdle(baddie);
  inGame.activeWorld.baddies.push(baddie);

  return baddie;
}

function updateBaddies(dt) {
  const { baxton } = inGame;
  const { baddies, npcs } = inGame.activeWorld;

  let attackingBaddies = 0;

  for (const baddie of baddies) {
    if (inGame.gameOver) {
      if (baddie.state !== BaddieStates.Idle) setBaddieIdle(baddie);
      continue;
    }

    /**
     * @type {EntityLike[]}
     */
    const availableTargets = [];
    for (let npc of npcs) {
      if (npc.following) availableTargets.push(npc);
    }

    availableTargets.push(baxton);

    const chaseRange = 6;
    const stopChaseRange = 10;

    switch (baddie.state) {
      case BaddieStates.Idle: {
        baddie.stateTimer -= dt;

        if (baddie.stateTimer <= 0) {
          for (let target of availableTargets) {
            const diff = { x: target.pos.x - baddie.pos.x, y: target.pos.y - baddie.pos.y };
            const dist = lengthxy(diff);

            if (dist < chaseRange) {
              setBaddieChaseTarget(baddie, target);
              break;
            }
          }
        }
        break;
      }
      case BaddieStates.Wander: {
        break;
      }
      case BaddieStates.ChaseTarget: {
        attackingBaddies++;

        if (availableTargets.indexOf(baddie.stateTarget) === -1) {
          setBaddieIdle(baddie);
          break;
        }

        const diff = { x: baddie.stateTarget.pos.x - baddie.pos.x, y: baddie.stateTarget.pos.y - baddie.pos.y };
        const angle = Math.atan2(diff.y, diff.x);
        const dist = lengthxy(diff);

        const attackRange = (baddie.collideBox.width + baddie.stateTarget.collideBox.width);

        if (dist < attackRange) {
          setBaddieAttack(baddie);
        } else if (dist > stopChaseRange) {
          setBaddieIdle(baddie);
        } else {
          const amount = Math.min(dist - baddie.collideBox.width + baddie.stateTarget.collideBox.width, baddie.speed * dt);
          moveWithCollision(baddie, { x: Math.cos(angle) * amount, y: Math.sin(angle) * amount });
        }

        break;
      }
      case BaddieStates.AttackTarget: {
        attackingBaddies++;

        if (availableTargets.indexOf(baddie.stateTarget) === -1) {
          setBaddieIdle(baddie);
          break;
        }

        const diff = { x: baddie.stateTarget.pos.x - baddie.pos.x, y: baddie.stateTarget.pos.y - baddie.pos.y };
        const dist = lengthxy(diff);

        const attackRange = (baddie.collideBox.width + baddie.stateTarget.collideBox.width);
        const hitRange = attackRange * 1.2;

        baddie.stateTimer -= dt;
        if (baddie.stateTimer <= 0) {
          if (dist < hitRange) {
            const attackAngle = anglexy(baddie.pos, baddie.stateTarget.pos);

            if (baddie.stateTarget == baxton) {
              damageBaxton(1, attackAngle);
            }
            else {
              /** @type {any} */
              const npcTarget = baddie.stateTarget;
              damageNpc(npcTarget, 1, attackAngle);
            }
          }

          setBaddieIdle(baddie);
        }
        break;
      }
      case BaddieStates.Hit: {
        baddie.stateTimer -= dt;
        if (baddie.stateTimer <= 0) {
          baddie.bodySprite.loop = true;
          setBaddieIdle(baddie);
        } else {
          moveWithCollision(baddie, baddie.knockbackDirection);
          baddie.knockbackDirection.x *= 0.9;
          baddie.knockbackDirection.y *= 0.9;
        }
      }
    }
  }

  // Combat music
  const isNowInCombat = attackingBaddies > 0;

  if (isNowInCombat) {
    combatCountdown = 5;

    if (!isInCombat) {
      isInCombat = true;
      updateCombatBgmVolume();
    }
  } else if (isInCombat) {
    combatCountdown -= dt;

    if (combatCountdown <= 0) {
      isInCombat = false;
      updateCombatBgmVolume();
    }
  }

  // Pushaway
  for (const baddie of baddies) {
    for (const otherEnemy of baddies) {
      if (baddie !== otherEnemy) pushaway(baddie, otherEnemy, dt);
    }

    pushaway(baddie, baxton, dt);
  }

  for (const baddie of baddies) {
    baddie.pixiRoot.position.copyFrom(baddie.pos);
    baddie.pixiRoot.zIndex = baddie.pos.y;

    baddie.shadowSprite.position.copyFrom(baddie.pos);
    baddie.shadowSprite.zIndex = baddie.pos.y - 10;

    baddie.bodySprite.update(app.ticker.deltaTime);
  }
}

/**
 * @param {Baddie} baddie
 */
function setBaddieIdle(baddie) {
  baddie.state = BaddieStates.Idle;
  baddie.stateTimer = 3;

  setAnimation(baddie.bodySprite, baddie.bodySpritesheet.animations.idle);
}

/**
 * @param {Baddie} baddie
 * @param {EntityLike} target
 */
function setBaddieChaseTarget(baddie, target) {
  baddie.state = BaddieStates.ChaseTarget;
  baddie.stateTarget = target;

  setAnimation(baddie.bodySprite, baddie.bodySpritesheet.animations.run);
}

/**
 * @param {Baddie} baddie
 */
function setBaddieAttack(baddie) {
  baddie.state = BaddieStates.AttackTarget;
  baddie.stateTimer = 0.3;

  setAnimation(baddie.bodySprite, baddie.bodySpritesheet.animations.attack);
}

/**
 * @param {Baddie} baddie
 */
function setBaddieHit(baddie, angle) {
  baddie.state = BaddieStates.Hit;
  baddie.stateTimer = 0.3;

  baddie.knockbackDirection.set(-Math.cos(angle) * 0.05, -Math.sin(angle) * 0.05);

  baddie.bodySprite.loop = false;
  setAnimation(baddie.bodySprite, baddie.bodySpritesheet.animations.hit);
}

/**
 * @param {Baddie} baddie
 * @param {number} damage
 * @param {number} angle
 */
function damageBaddie(baddie, damage, angle) {
  baddie.health = Math.max(baddie.health - damage, 0);

  play2Dsound(`${baddie.health === 0 ? "deaths" : "hits"}.baddies.${baddie.baddieDesc.id}`, baddie.pos);

  if (baddie.health === 0) {
    inGame.activeWorld.pixiLayers["ground"].entitiesContainer.removeChild(baddie.pixiRoot);
    inGame.activeWorld.pixiShadowsLayer.removeChild(baddie.shadowSprite);
    inGame.activeWorld.baddies.splice(inGame.activeWorld.baddies.indexOf(baddie), 1);

    baddie.bodySprite.destroy();
    baddie.shadowSprite.destroy();

    spawnAnimatedFx("inGame/fxs/vanish", baddie.pos, 0);

    if (randInt(0, 3) === 0) {
      const type = randInt(0, 3) === 0 ? "armor" : "heart";
      getSound(`pickUp.${type}`).play();

      if (type === "heart") inGame.stats.health = Math.min(inGame.stats.maxHealth, inGame.stats.health + 1);
      else inGame.stats.armor = Math.min(inGame.stats.maxArmor, inGame.stats.armor + 1);
      updateHudHealth();

      let timer = 0;
      const duration = 0.5;

      spawnFx(`inGame/fxs/${type}PickUp`, baddie.pos, 0, (fx, dt) => {
        timer = Math.min(duration, timer + dt);
        const newPos = lerpxy(fx.sprite.position, inGame.baxton.pos, timer / duration);
        fx.sprite.position.set(newPos.x, newPos.y);
        return timer < duration;
      });
    }
  } else {
    setBaddieHit(baddie, angle);
  }
}
